@@ -47,6 +47,7 @@
 #include "imports.h"
 #include "transform.h"
 
+#define WITH_XSLT_DEBUG_ATTRIBUTES
 #ifdef WITH_XSLT_DEBUG
 #define WITH_XSLT_DEBUG_ATTRIBUTES
 #endif
@@ -122,10 +123,6 @@ xsltNewAttrElem(xmlNodePtr attr) {
  */
 static void
 xsltFreeAttrElem(xsltAttrElemPtr attr) {
-    if (attr->set != NULL)
-	xmlFree((char *)attr->set);
-    if (attr->ns != NULL)
-	xmlFree((char *)attr->ns);
     xmlFree(attr);
 }
 
@@ -170,6 +167,7 @@ xsltAddAttrElemList(xsltAttrElemPtr list, xmlNodePtr attr) {
 	    return(cur);
 	if (cur->next == NULL) {
 	    cur->next = xsltNewAttrElem(attr);
+	    return(list);
 	}
 	cur = next;
     }
@@ -266,19 +264,18 @@ xsltMergeAttrElemList(xsltAttrElemPtr list, xsltAttrElemPtr old) {
 /**
  * xsltParseStylesheetAttributeSet:
  * @style:  the XSLT stylesheet
- * @template:  the "preserve-space" element
+ * @cur:  the "attribute-set" element
  *
- * parse an XSLT stylesheet preserve-space element and record
- * elements needing preserving
+ * parse an XSLT stylesheet attribute-set element
  */
 
 void
 xsltParseStylesheetAttributeSet(xsltStylesheetPtr style, xmlNodePtr cur) {
-    xmlChar *prop = NULL;
-    xmlChar *ncname = NULL;
-    xmlChar *prefix = NULL;
+    const xmlChar *ncname;
+    const xmlChar *prefix;
+    const xmlChar *attrib, *endattr;
+    xmlChar *prop;
     xmlChar *attributes;
-    xmlChar *attrib, *endattr;
     xmlNodePtr list;
     xsltAttrElemPtr values;
 
@@ -289,15 +286,11 @@ xsltParseStylesheetAttributeSet(xsltStylesheetPtr style, xmlNodePtr cur) {
     if (prop == NULL) {
 	xsltGenericError(xsltGenericErrorContext,
 	     "xsl:attribute-set : name is missing\n");
-	goto error;
+	return;
     }
 
-    ncname = xmlSplitQName2(prop, &prefix);
-    if (ncname == NULL) {
-	ncname = prop;
-	prop = NULL;
-	prefix = NULL;
-    }
+    ncname = xsltSplitQName(style->dict, prop, &prefix);
+    xmlFree(prop);
 
     if (style->attributeSets == NULL) {
 #ifdef WITH_XSLT_DEBUG_ATTRIBUTES
@@ -307,7 +300,7 @@ xsltParseStylesheetAttributeSet(xsltStylesheetPtr style, xmlNodePtr cur) {
 	style->attributeSets = xmlHashCreate(10);
     }
     if (style->attributeSets == NULL)
-	goto error;
+	return;
 
     values = xmlHashLookup2(style->attributeSets, ncname, prefix);
 
@@ -353,34 +346,24 @@ xsltParseStylesheetAttributeSet(xsltStylesheetPtr style, xmlNodePtr cur) {
 	    break;
         endattr = attrib;
 	while ((*endattr != 0) && (!IS_BLANK(*endattr))) endattr++;
-	attrib = xmlStrndup(attrib, endattr - attrib);
+	attrib = xmlDictLookup(style->dict, attrib, endattr - attrib);
 	if (attrib) {
-	    xmlChar *ncname2 = NULL;
-	    xmlChar *prefix2 = NULL;
+	    const xmlChar *ncname2 = NULL;
+	    const xmlChar *prefix2 = NULL;
 	    xsltAttrElemPtr values2;
+
 #ifdef WITH_XSLT_DEBUG_ATTRIBUTES
 	    xsltGenericDebug(xsltGenericDebugContext,
 		"xsl:attribute-set : %s adds use %s\n", ncname, attrib);
 #endif
-	    ncname2 = xmlSplitQName2(attrib, &prefix2);
-	    if (ncname2 == NULL) {
-		ncname2 = attrib;
-		attrib = NULL;
-		prefix = NULL;
-	    }
+	    ncname2 = xsltSplitQName(style->dict, attrib, &prefix2);
 	    values2 = xsltNewAttrElem(NULL);
 	    if (values2 != NULL) {
 		values2->set = ncname2;
 		values2->ns = prefix2;
 		values = xsltMergeAttrElemList(values, values2);
+		xsltFreeAttrElem(values2);
 	    }
-
-	    if (attrib != NULL)
-		xmlFree(attrib);
-	    if (ncname2 != NULL)
-		xmlFree(ncname2);
-	    if (prefix2 != NULL)
-		xmlFree(prefix2);
 	}
 	attrib = endattr;
     }
@@ -397,14 +380,6 @@ done:
     xsltGenericDebug(xsltGenericDebugContext,
 	"updated attribute list %s\n", ncname);
 #endif
-
-error:
-    if (prop != NULL)
-        xmlFree(prop);
-    if (ncname != NULL)
-        xmlFree(ncname);
-    if (prefix != NULL)
-        xmlFree(prefix);
 }
 
 /**
@@ -488,6 +463,48 @@ xsltResolveSASCallback(xsltAttrElemPtr values, xsltStylesheetPtr style,
     }
 }
 
+/**
+ * xsltMergeSASCallback,:
+ * @style:  the XSLT stylesheet
+ *
+ * Merge an attribute set from an imported stylesheet.
+ */
+static void
+xsltMergeSASCallback(xsltAttrElemPtr values, xsltStylesheetPtr style,
+	               const xmlChar *name, const xmlChar *ns,
+		       ATTRIBUTE_UNUSED const xmlChar *ignored) {
+    int ret;
+    xsltAttrElemPtr topSet;
+
+    ret = xmlHashAddEntry2(style->attributeSets, name, ns, values);
+    if (ret < 0) {
+	/*
+	 * Add failed, this attribute set can be removed.
+	 */
+#ifdef WITH_XSLT_DEBUG_ATTRIBUTES
+	xsltGenericDebug(xsltGenericDebugContext,
+		"attribute set %s present already in top stylesheet"
+		" - merging\n", name);
+#endif
+	topSet = xmlHashLookup2(style->attributeSets, name, ns);
+	if (topSet==NULL) {
+	    xsltGenericError(xsltGenericErrorContext,
+	        "xsl:attribute-set : logic error merging from imports for"
+		" attribute-set %s\n", name);
+	} else {
+	    topSet = xsltMergeAttrElemList(topSet, values);
+	    xmlHashUpdateEntry2(style->attributeSets, name, ns, topSet, NULL);
+	}
+	xsltFreeAttrElemList(values);
+#ifdef WITH_XSLT_DEBUG_ATTRIBUTES
+    } else {
+	xsltGenericDebug(xsltGenericDebugContext,
+		"attribute set %s moved to top stylesheet\n",
+		         name);
+#endif
+    }
+}
+
 /**
  * xsltResolveStylesheetAttributeSet:
  * @style:  the XSLT stylesheet
@@ -496,10 +513,40 @@ xsltResolveSASCallback(xsltAttrElemPtr values, xsltStylesheetPtr style,
  */
 void
 xsltResolveStylesheetAttributeSet(xsltStylesheetPtr style) {
+    xsltStylesheetPtr cur;
+
 #ifdef WITH_XSLT_DEBUG_ATTRIBUTES
     xsltGenericDebug(xsltGenericDebugContext,
 	    "Resolving attribute sets references\n");
 #endif
+    /*
+     * First aggregate all the attribute sets definitions from the imports
+     */
+    cur = xsltNextImport(style);
+    while (cur != NULL) {
+	if (cur->attributeSets != NULL) {
+	    if (style->attributeSets == NULL) {
+#ifdef WITH_XSLT_DEBUG_ATTRIBUTES
+		xsltGenericDebug(xsltGenericDebugContext,
+		    "creating attribute set table\n");
+#endif
+		style->attributeSets = xmlHashCreate(10);
+	    }
+	    xmlHashScanFull(cur->attributeSets, 
+		(xmlHashScannerFull) xsltMergeSASCallback, style);
+	    /*
+	     * the attribute lists have either been migrated to style
+	     * or freed directly in xsltMergeSASCallback()
+	     */
+	    xmlHashFree(cur->attributeSets, NULL);
+	    cur->attributeSets = NULL;
+	}
+	cur = xsltNextImport(cur);
+    }
+
+    /*
+     * Then resolve all the references and computes the resulting sets
+     */
     if (style->attributeSets != NULL) {
 	xmlHashScanFull(style->attributeSets, 
 		(xmlHashScannerFull) xsltResolveSASCallback, style);
@@ -522,8 +569,9 @@ xsltAttributeInternal(xsltTransformContextPtr ctxt, xmlNodePtr node,
                       int fromset)
 {
     xmlChar *prop = NULL;
-    xmlChar *ncname = NULL, *name, *namespace;
-    xmlChar *prefix = NULL;
+    xmlChar *namespace;
+    const xmlChar *name = NULL;
+    const xmlChar *prefix = NULL;
     xmlChar *value = NULL;
     xmlNsPtr ns = NULL;
     xmlAttrPtr attr;
@@ -550,7 +598,7 @@ xsltAttributeInternal(xsltTransformContextPtr ctxt, xmlNodePtr node,
         return;
     }
 #ifdef WITH_DEBUGGER
-    if (xslDebugStatus != XSLT_DEBUG_NONE) {
+    if (ctxt->debugStatus != XSLT_DEBUG_NONE) {
         xslHandleDebugger(inst, node, NULL, ctxt);
     }
 #endif
@@ -564,21 +612,21 @@ xsltAttributeInternal(xsltTransformContextPtr ctxt, xmlNodePtr node,
                              "xsl:attribute : name is missing\n");
             goto error;
         }
-        name = prop;
+	if (xmlValidateQName(prop, 0)) {
+	    xsltTransformError(ctxt, NULL, inst,
+			    "xsl:attribute : invalid QName\n");
+	    /* we fall through to catch any further errors, if possible */
+	}
+	name = xsltSplitQName(ctxt->dict, prop, &prefix);
+	xmlFree(prop);
     } else {
-        name = comp->name;
+	name = xsltSplitQName(ctxt->dict, comp->name, &prefix);
     }
 
-    ncname = xmlSplitQName2(name, &prefix);
-    if (ncname == NULL) {
-        prefix = NULL;
-    } else {
-        name = ncname;
-    }
-    if (!xmlStrncasecmp(prefix, (xmlChar *) "xml", 3)) {
+    if (!xmlStrncasecmp(prefix, (xmlChar *) "xmlns", 5)) {
 #ifdef WITH_XSLT_DEBUG_PARSING
         xsltGenericDebug(xsltGenericDebugContext,
-                         "xsltAttribute: xml prefix forbidden\n");
+                         "xsltAttribute: xmlns prefix forbidden\n");
 #endif
         goto error;
     }
@@ -617,7 +665,10 @@ xsltAttributeInternal(xsltTransformContextPtr ctxt, xmlNodePtr node,
         URL = ns->href;
 
     if (fromset) {
-	attr = xmlHasNsProp(ctxt->insert, name, ns);
+	if (URL != NULL)
+	    attr = xmlHasNsProp(ctxt->insert, name, URL);
+	else
+	    attr = xmlHasProp(ctxt->insert, name);
 	if (attr != NULL)
 	    return;
     }
@@ -638,13 +689,7 @@ xsltAttributeInternal(xsltTransformContextPtr ctxt, xmlNodePtr node,
         }
     }
 
-  error:
-    if (prop != NULL)
-        xmlFree(prop);
-    if (ncname != NULL)
-        xmlFree(ncname);
-    if (prefix != NULL)
-        xmlFree(prefix);
+error:
     if (value != NULL)
         xmlFree(value);
 }
@@ -677,11 +722,11 @@ xsltAttribute(xsltTransformContextPtr ctxt, xmlNodePtr node,
 void
 xsltApplyAttributeSet(xsltTransformContextPtr ctxt, xmlNodePtr node,
                       xmlNodePtr inst ATTRIBUTE_UNUSED,
-                      xmlChar * attributes)
+                      const xmlChar * attributes)
 {
-    xmlChar *ncname = NULL;
-    xmlChar *prefix = NULL;
-    xmlChar *attrib, *endattr;
+    const xmlChar *ncname = NULL;
+    const xmlChar *prefix = NULL;
+    const xmlChar *attrib, *endattr;
     xsltAttrElemPtr values;
     xsltStylesheetPtr style;
 
@@ -698,23 +743,18 @@ xsltApplyAttributeSet(xsltTransformContextPtr ctxt, xmlNodePtr node,
         endattr = attrib;
         while ((*endattr != 0) && (!IS_BLANK(*endattr)))
             endattr++;
-        attrib = xmlStrndup(attrib, endattr - attrib);
+        attrib = xmlDictLookup(ctxt->dict, attrib, endattr - attrib);
         if (attrib) {
 #ifdef WITH_XSLT_DEBUG_ATTRIBUTES
             xsltGenericDebug(xsltGenericDebugContext,
                              "apply attribute set %s\n", attrib);
 #endif
-            ncname = xmlSplitQName2(attrib, &prefix);
-            if (ncname == NULL) {
-                ncname = attrib;
-                attrib = NULL;
-                prefix = NULL;
-            }
+            ncname = xsltSplitQName(ctxt->dict, attrib, &prefix);
 
             style = ctxt->style;
 #ifdef WITH_DEBUGGER
             if ((style != NULL) && (style->attributeSets != NULL) &&
-		(xslDebugStatus != XSLT_DEBUG_NONE)) {
+		(ctxt->debugStatus != XSLT_DEBUG_NONE)) {
                 values =
                     xmlHashLookup2(style->attributeSets, ncname, prefix);
                 if ((values != NULL) && (values->attr != NULL))
@@ -734,12 +774,6 @@ xsltApplyAttributeSet(xsltTransformContextPtr ctxt, xmlNodePtr node,
                 }
                 style = xsltNextImport(style);
             }
-            if (attrib != NULL)
-                xmlFree(attrib);
-            if (ncname != NULL)
-                xmlFree(ncname);
-            if (prefix != NULL)
-                xmlFree(prefix);
         }
         attrib = endattr;
     }
