@@ -43,6 +43,8 @@ struct _xsltKeyDef {
     xmlChar *use;
     xmlXPathCompExprPtr comp;
     xmlXPathCompExprPtr usecomp;
+    xmlNsPtr *nsList;           /* the namespaces in scope */
+    int nsNr;                   /* the number of namespaces in scope */
 };
 
 typedef struct _xsltKeyTable xsltKeyTable;
@@ -76,8 +78,7 @@ xsltNewKeyDef(const xmlChar *name, const xmlChar *nameURI) {
 
     cur = (xsltKeyDefPtr) xmlMalloc(sizeof(xsltKeyDef));
     if (cur == NULL) {
-	xsltPrintErrorContext(NULL, NULL, NULL);
-        xsltGenericError(xsltGenericErrorContext,
+	xsltTransformError(NULL, NULL, NULL,
 		"xsltNewKeyDef : malloc failed\n");
 	return(NULL);
     }
@@ -86,6 +87,7 @@ xsltNewKeyDef(const xmlChar *name, const xmlChar *nameURI) {
 	cur->name = xmlStrdup(name);
     if (nameURI != NULL)
 	cur->nameURI = xmlStrdup(nameURI);
+    cur->nsList = NULL;
     return(cur);
 }
 
@@ -111,6 +113,8 @@ xsltFreeKeyDef(xsltKeyDefPtr keyd) {
 	xmlFree(keyd->match);
     if (keyd->use != NULL)
 	xmlFree(keyd->use);
+    if (keyd->nsList != NULL)
+        xmlFree(keyd->nsList);
     memset(keyd, -1, sizeof(xsltKeyDef));
     xmlFree(keyd);
 }
@@ -147,8 +151,7 @@ xsltNewKeyTable(const xmlChar *name, const xmlChar *nameURI) {
 
     cur = (xsltKeyTablePtr) xmlMalloc(sizeof(xsltKeyTable));
     if (cur == NULL) {
-	xsltPrintErrorContext(NULL, NULL, NULL);
-        xsltGenericError(xsltGenericErrorContext,
+	xsltTransformError(NULL, NULL, NULL,
 		"xsltNewKeyTable : malloc failed\n");
 	return(NULL);
     }
@@ -218,6 +221,63 @@ xsltFreeKeys(xsltStylesheetPtr style) {
 	xsltFreeKeyDefList((xsltKeyDefPtr) style->keys);
 }
 
+/**
+ * skipString:
+ * @cur: the current pointer
+ * @end: the current offset
+ *
+ * skip a string delimited by " or '
+ *
+ * Returns the byte after the string or -1 in case of error
+ */
+static int
+skipString(const xmlChar *cur, int end) {
+    xmlChar limit;
+
+    if ((cur == NULL) || (end < 0)) return(-1);
+    if ((cur[end] == '\'') || (cur[end] == '"')) limit = cur[end];
+    else return(end);
+    end++;
+    while (cur[end] != 0) {
+        if (cur[end] == limit)
+	    return(end + 1);
+	end++;
+    }
+    return(-1);
+}
+
+/**
+ * skipPredicate:
+ * @cur: the current pointer
+ * @end: the current offset
+ *
+ * skip a predicate
+ *
+ * Returns the byte after the predicate or -1 in case of error
+ */
+static int
+skipPredicate(const xmlChar *cur, int end) {
+    if ((cur == NULL) || (end < 0)) return(-1);
+    if (cur[end] != '[') return(end);
+    end++;
+    while (cur[end] != 0) {
+        if ((cur[end] == '\'') || (cur[end] == '"')) {
+	    end = skipString(cur, end);
+	    if (end <= 0)
+	        return(-1);
+	    continue;
+	} else if (cur[end] == '[') {
+	    end = skipPredicate(cur, end);
+	    if (end <= 0)
+	        return(-1);
+	    continue;
+	} else if (cur[end] == ']')
+	    return(end + 1);
+	end++;
+    }
+    return(-1);
+}
+
 /**
  * xsltAddKey:
  * @style: an XSLT stylesheet
@@ -237,7 +297,7 @@ xsltAddKey(xsltStylesheetPtr style, const xmlChar *name,
 	   const xmlChar *use, xmlNodePtr inst) {
     xsltKeyDefPtr key;
     xmlChar *pattern = NULL;
-    int current, end, start;
+    int current, end, start, i = 0;
 
     if ((style == NULL) || (name == NULL) || (match == NULL) || (use == NULL))
 	return(-1);
@@ -251,6 +311,12 @@ xsltAddKey(xsltStylesheetPtr style, const xmlChar *name,
     key->match = xmlStrdup(match);
     key->use = xmlStrdup(use);
     key->inst = inst;
+    key->nsList = xmlGetNsList(inst->doc, inst);
+    if (key->nsList != NULL) {
+        while (key->nsList[i] != NULL)
+	    i++;
+    }
+    key->nsNr = i;
 
     /*
      * Split the | and register it as as many keys
@@ -258,29 +324,38 @@ xsltAddKey(xsltStylesheetPtr style, const xmlChar *name,
     current = end = 0;
     while (match[current] != 0) {
 	start = current;
-	while (IS_BLANK(match[current]))
+	while (IS_BLANK_CH(match[current]))
 	    current++;
 	end = current;
 	while ((match[end] != 0) && (match[end] != '|')) {
-	    end++;
+	    if (match[end] == '[') {
+	        end = skipPredicate(match, end);
+		if (end <= 0) {
+		    xsltTransformError(NULL, style, inst,
+		                       "key pattern is malformed: %s",
+				       key->match);
+		    if (style != NULL) style->errors++;
+		    goto error;
+		}
+	    } else
+		end++;
 	}
 	if (current == end) {
-	    xsltPrintErrorContext(NULL, style, inst);
-	    xsltGenericError(xsltGenericErrorContext,
-			     "key pattern is empty\n");
-	    style->errors++;
+	    xsltTransformError(NULL, style, inst,
+			       "key pattern is empty\n");
+	    if (style != NULL) style->errors++;
 	    goto error;
 	}
 	if (match[start] != '/') {
 	    pattern = xmlStrcat(pattern, (xmlChar *)"//");
 	    if (pattern == NULL) {
-		style->errors++;
+		if (style != NULL) style->errors++;
 		goto error;
 	    }
 	}
 	pattern = xmlStrncat(pattern, &match[start], end - start);
 	if (pattern == NULL) {
-	    style->errors++;
+	    if (style != NULL) style->errors++;
 	    goto error;
 	}
 
@@ -294,21 +369,19 @@ xsltAddKey(xsltStylesheetPtr style, const xmlChar *name,
     xsltGenericDebug(xsltGenericDebugContext,
 	"   resulting pattern %s\n", pattern);
 #endif
-    key->comp = xmlXPathCompile(pattern);
+    key->comp = xsltXPathCompile(style, pattern);
     if (key->comp == NULL) {
-	xsltPrintErrorContext(NULL, style, inst);
-        xsltGenericError(xsltGenericErrorContext,
+	xsltTransformError(NULL, style, inst,
 		"xsl:key : XPath pattern compilation failed '%s'\n",
 		         pattern);
-	style->errors++;
+	if (style != NULL) style->errors++;
     }
-    key->usecomp = xmlXPathCompile(use);
+    key->usecomp = xsltXPathCompile(style, use);
     if (key->usecomp == NULL) {
-	xsltPrintErrorContext(NULL, style, inst);
-        xsltGenericError(xsltGenericErrorContext,
+	xsltTransformError(NULL, style, inst,
 		"xsl:key : XPath pattern compilation failed '%s'\n",
 		         use);
-	style->errors++;
+	if (style != NULL) style->errors++;
     }
     key->next = style->keys;
     style->keys = key;
@@ -335,7 +408,8 @@ xsltGetKey(xsltTransformContextPtr ctxt, const xmlChar *name,
     xmlNodeSetPtr ret;
     xsltKeyTablePtr table;
 
-    if ((ctxt == NULL) || (name == NULL) || (value == NULL))
+    if ((ctxt == NULL) || (name == NULL) || (value == NULL) ||
+	(ctxt->document == NULL))
 	return(NULL);
 
 #ifdef WITH_XSLT_DEBUG_KEYS
@@ -367,7 +441,8 @@ xsltGetKey(xsltTransformContextPtr ctxt, const xmlChar *name,
  *         by the caller.
  */
 static xmlChar **
-xsltEvalXPathKeys(xsltTransformContextPtr ctxt, xmlXPathCompExprPtr comp) {
+xsltEvalXPathKeys(xsltTransformContextPtr ctxt, xmlXPathCompExprPtr comp,
+                  xsltKeyDefPtr keyd) {
     xmlChar **ret = NULL;
     xmlXPathObjectPtr res;
     xmlNodePtr oldInst;
@@ -384,9 +459,8 @@ xsltEvalXPathKeys(xsltTransformContextPtr ctxt, xmlXPathCompExprPtr comp) {
     oldNamespaces = ctxt->xpathCtxt->namespaces;
 
     ctxt->xpathCtxt->node = ctxt->node;
-    /* TODO: do we need to propagate the namespaces here ? */
-    ctxt->xpathCtxt->namespaces = NULL;
-    ctxt->xpathCtxt->nsNr = 0;
+    ctxt->xpathCtxt->namespaces = keyd->nsList;
+    ctxt->xpathCtxt->nsNr = keyd->nsNr;
     res = xmlXPathCompiledEval(comp, ctxt->xpathCtxt);
     if (res != NULL) {
 	if (res->type == XPATH_NODESET) {
@@ -394,6 +468,8 @@ xsltEvalXPathKeys(xsltTransformContextPtr ctxt, xmlXPathCompExprPtr comp) {
 
 	    if (res->nodesetval != NULL)
 		len = res->nodesetval->nodeNr;
+	    else
+		len = 0;
 	    if (len != 0) {
 		ret = (xmlChar **) xmlMalloc((len + 1) * sizeof(xmlChar *));
 		if (ret != NULL) {
@@ -417,8 +493,7 @@ xsltEvalXPathKeys(xsltTransformContextPtr ctxt, xmlXPathCompExprPtr comp) {
 		    res->stringval = NULL;
 		}
 	    } else {
-		xsltPrintErrorContext(ctxt, NULL, NULL);
-		xsltGenericError(xsltGenericErrorContext,
+		xsltTransformError(ctxt, NULL, NULL,
 		     "xpath : string() function didn't return a String\n");
 	    }
 	}
@@ -496,21 +571,21 @@ xsltInitCtxtKey(xsltTransformContextPtr ctxt, xsltDocumentPtr doc,
 	    nodelist = res->nodesetval;
 #ifdef WITH_XSLT_DEBUG_KEYS
 	    if (nodelist != NULL)
-		xsltGenericDebug(xsltGenericDebugContext,
+		XSLT_TRACE(ctxt,XSLT_TRACE_KEYS,xsltGenericDebug(xsltGenericDebugContext,
 		     "xsltInitCtxtKey: %s evaluates to %d nodes\n",
-				 keyd->match, nodelist->nodeNr);
+				 keyd->match, nodelist->nodeNr));
 #endif
 	} else {
 #ifdef WITH_XSLT_DEBUG_KEYS
-	    xsltGenericDebug(xsltGenericDebugContext,
-		 "xsltInitCtxtKey: %s is not a node set\n", keyd->match);
+	    XSLT_TRACE(ctxt,XSLT_TRACE_KEYS,xsltGenericDebug(xsltGenericDebugContext,
+		 "xsltInitCtxtKey: %s is not a node set\n", keyd->match));
 #endif
 	    goto error;
 	}
     } else {
 #ifdef WITH_XSLT_DEBUG_KEYS
-	xsltGenericDebug(xsltGenericDebugContext,
-	     "xsltInitCtxtKey: %s evaluation failed\n", keyd->match);
+	XSLT_TRACE(ctxt,XSLT_TRACE_KEYS,xsltGenericDebug(xsltGenericDebugContext,
+	     "xsltInitCtxtKey: %s evaluation failed\n", keyd->match));
 #endif
 	ctxt->state = XSLT_STATE_STOPPED;
 	goto error;
@@ -522,24 +597,45 @@ xsltInitCtxtKey(xsltTransformContextPtr ctxt, xsltDocumentPtr doc,
     if ((nodelist == NULL) || (nodelist->nodeNr <= 0))
 	goto error;
 
-    table = xsltNewKeyTable(keyd->name, keyd->nameURI);
-    if (table == NULL)
-	goto error;
+    /**
+     * Multiple key definitions for the same name are allowed, so
+     * we must check if the key is already present for this doc
+     */
+    table = (xsltKeyTablePtr) doc->keys;
+    while (table != NULL) {
+        if (xmlStrEqual(table->name, keyd->name) &&
+	    (((keyd->nameURI == NULL) && (table->nameURI == NULL)) ||
+	     ((keyd->nameURI != NULL) && (table->nameURI != NULL) &&
+	      (xmlStrEqual(table->nameURI, keyd->nameURI)))))
+	    break;
+	table = table->next;
+    }
+    /**
+     * If the key was not previously defined, create it now and
+     * chain it to the list of keys for the doc
+     */
+    if (table == NULL) {
+        table = xsltNewKeyTable(keyd->name, keyd->nameURI);
+        if (table == NULL)
+	    goto error;
+        table->next = doc->keys;
+        doc->keys = table;
+    }
 
     for (i = 0;i < nodelist->nodeNr;i++) {
 	if (IS_XSLT_REAL_NODE(nodelist->nodeTab[i])) {
 	    ctxt->node = nodelist->nodeTab[i];
 
-	    list = xsltEvalXPathKeys(ctxt, keyd->usecomp);
+	    list = xsltEvalXPathKeys(ctxt, keyd->usecomp, keyd);
 	    if (list != NULL) {
 		int index = 0;
 
 		str = list[index++];
 		while (str != NULL) {
 #ifdef WITH_XSLT_DEBUG_KEYS
-		    xsltGenericDebug(xsltGenericDebugContext,
+		    XSLT_TRACE(ctxt,XSLT_TRACE_KEYS,xsltGenericDebug(xsltGenericDebugContext,
 			 "xsl:key : node associated to(%s,%s)\n",
-				     keyd->name, str);
+				     keyd->name, str));
 #endif
 		    keylist = xmlHashLookup(table->keys, str);
 		    if (keylist == NULL) {
@@ -555,17 +651,14 @@ xsltInitCtxtKey(xsltTransformContextPtr ctxt, xsltDocumentPtr doc,
 		xmlFree(list);
 #ifdef WITH_XSLT_DEBUG_KEYS
 	    } else {
-		xsltGenericDebug(xsltGenericDebugContext,
+		XSLT_TRACE(ctxt,XSLT_TRACE_KEYS,xsltGenericDebug(xsltGenericDebugContext,
 		     "xsl:key : use %s failed to return strings\n",
-				 keyd->use);
+				 keyd->use));
 #endif
 	    }
 	}
     }
 
-    table->next = doc->keys;
-    doc->keys = table;
-
 error:
     ctxt->document = oldDoc;
     ctxt->xpathCtxt->doc = oldXDoc;
@@ -592,8 +685,8 @@ xsltInitCtxtKeys(xsltTransformContextPtr ctxt, xsltDocumentPtr doc) {
 	return;
 #ifdef WITH_XSLT_DEBUG_KEYS
     if ((doc->doc != NULL) && (doc->doc->URL != NULL))
-	xsltGenericDebug(xsltGenericDebugContext, "Initializing keys on %s\n",
-		     doc->doc->URL);
+	XSLT_TRACE(ctxt,XSLT_TRACE_KEYS,xsltGenericDebug(xsltGenericDebugContext, "Initializing keys on %s\n",
+		     doc->doc->URL));
 #endif
     style = ctxt->style;
     while (style != NULL) {
