@@ -39,6 +39,18 @@ libxslt_xsltParseStylesheetImportedDoc(PyObject *self ATTRIBUTE_UNUSED, PyObject
     return(py_retval);
 }
 
+PyObject *
+libxslt_xsltCalibrateAdjust(PyObject *self ATTRIBUTE_UNUSED, PyObject *args) {
+    long delta;
+
+    if (!PyArg_ParseTuple(args, (char *)"i:xsltCalibrateAdjust", &delta))
+        return(NULL);
+
+    xsltCalibrateAdjust(delta);
+    Py_INCREF(Py_None);
+    return(Py_None);
+}
+
 PyObject *
 libxslt_xsltGenerateIdFunction(PyObject *self ATTRIBUTE_UNUSED, PyObject *args) {
     xmlXPathParserContextPtr ctxt;
@@ -885,25 +897,6 @@ libxslt_xsltApplyAttributeSet(PyObject *self ATTRIBUTE_UNUSED, PyObject *args) {
     return(Py_None);
 }
 
-PyObject *
-libxslt_xsltRegisterTmpRVT(PyObject *self ATTRIBUTE_UNUSED, PyObject *args) {
-    PyObject *py_retval;
-    int c_retval;
-    xsltTransformContextPtr ctxt;
-    PyObject *pyobj_ctxt;
-    xmlDocPtr RVT;
-    PyObject *pyobj_RVT;
-
-    if (!PyArg_ParseTuple(args, (char *)"OO:xsltRegisterTmpRVT", &pyobj_ctxt, &pyobj_RVT))
-        return(NULL);
-    ctxt = (xsltTransformContextPtr) PytransformCtxt_Get(pyobj_ctxt);
-    RVT = (xmlDocPtr) PyxmlNode_Get(pyobj_RVT);
-
-    c_retval = xsltRegisterTmpRVT(ctxt, RVT);
-    py_retval = libxml_intWrap((int) c_retval);
-    return(py_retval);
-}
-
 PyObject *
 libxslt_xsltUnregisterExtModuleTopLevel(PyObject *self ATTRIBUTE_UNUSED, PyObject *args) {
     PyObject *py_retval;
@@ -1186,6 +1179,22 @@ libxslt_xsltStylesheetGetEncoding(PyObject *self ATTRIBUTE_UNUSED, PyObject *arg
     return(py_retval);
 }
 
+PyObject *
+libxslt_xsltTransformGetStyle(PyObject *self ATTRIBUTE_UNUSED, PyObject *args) {
+    PyObject *py_retval;
+    xsltStylesheetPtr c_retval;
+    xsltTransformContextPtr ctxt;
+    PyObject *pyobj_ctxt;
+
+    if (!PyArg_ParseTuple(args, (char *)"O:xsltTransformGetStyle", &pyobj_ctxt))
+        return(NULL);
+    ctxt = (xsltTransformContextPtr) PytransformCtxt_Get(pyobj_ctxt);
+
+    c_retval = ctxt->style;
+    py_retval = libxslt_xsltStylesheetPtrWrap((xsltStylesheetPtr) c_retval);
+    return(py_retval);
+}
+
 PyObject *
 libxslt_xsltTransformGetMode(PyObject *self ATTRIBUTE_UNUSED, PyObject *args) {
     PyObject *py_retval;
@@ -1556,6 +1565,25 @@ libxslt_xsltTransformGetCurrent(PyObject *self ATTRIBUTE_UNUSED, PyObject *args)
     return(py_retval);
 }
 
+PyObject *
+libxslt_xsltRegisterTmpRVT(PyObject *self ATTRIBUTE_UNUSED, PyObject *args) {
+    PyObject *py_retval;
+    int c_retval;
+    xsltTransformContextPtr ctxt;
+    PyObject *pyobj_ctxt;
+    xmlDocPtr RVT;
+    PyObject *pyobj_RVT;
+
+    if (!PyArg_ParseTuple(args, (char *)"OO:xsltRegisterTmpRVT", &pyobj_ctxt, &pyobj_RVT))
+        return(NULL);
+    ctxt = (xsltTransformContextPtr) PytransformCtxt_Get(pyobj_ctxt);
+    RVT = (xmlDocPtr) PyxmlNode_Get(pyobj_RVT);
+
+    c_retval = xsltRegisterTmpRVT(ctxt, RVT);
+    py_retval = libxml_intWrap((int) c_retval);
+    return(py_retval);
+}
+
 PyObject *
 libxslt_xsltPrintErrorContext(PyObject *self ATTRIBUTE_UNUSED, PyObject *args) {
     xsltTransformContextPtr ctxt;
@@ -1592,6 +1620,27 @@ libxslt_xsltStylesheetGetDoc(PyObject *self ATTRIBUTE_UNUSED, PyObject *args) {
     return(py_retval);
 }
 
+PyObject *
+libxslt_xsltGetCNsProp(PyObject *self ATTRIBUTE_UNUSED, PyObject *args) {
+    PyObject *py_retval;
+    const xmlChar * c_retval;
+    xsltStylesheetPtr style;
+    PyObject *pyobj_style;
+    xmlNodePtr node;
+    PyObject *pyobj_node;
+    xmlChar * name;
+    xmlChar * nameSpace;
+
+    if (!PyArg_ParseTuple(args, (char *)"OOzz:xsltGetCNsProp", &pyobj_style, &pyobj_node, &name, &nameSpace))
+        return(NULL);
+    style = (xsltStylesheetPtr) Pystylesheet_Get(pyobj_style);
+    node = (xmlNodePtr) PyxmlNode_Get(pyobj_node);
+
+    c_retval = xsltGetCNsProp(style, node, name, nameSpace);
+    py_retval = libxml_xmlCharPtrConstWrap((const xmlChar *) c_retval);
+    return(py_retval);
+}
+
 PyObject *
 libxslt_xsltDebugDumpExtensions(PyObject *self ATTRIBUTE_UNUSED, PyObject *args) {
     FILE * output;
@@ -1717,19 +1766,20 @@ libxslt_xsltRegisterAllExtras(PyObject *self ATTRIBUTE_UNUSED, PyObject *args AT
 }
 
 PyObject *
-libxslt_xsltTransformGetStyle(PyObject *self ATTRIBUTE_UNUSED, PyObject *args) {
-    PyObject *py_retval;
-    xsltStylesheetPtr c_retval;
-    xsltTransformContextPtr ctxt;
-    PyObject *pyobj_ctxt;
+libxslt_xsltCompileAttr(PyObject *self ATTRIBUTE_UNUSED, PyObject *args) {
+    xsltStylesheetPtr style;
+    PyObject *pyobj_style;
+    xmlAttrPtr attr;
+    PyObject *pyobj_attr;
 
-    if (!PyArg_ParseTuple(args, (char *)"O:xsltTransformGetStyle", &pyobj_ctxt))
+    if (!PyArg_ParseTuple(args, (char *)"OO:xsltCompileAttr", &pyobj_style, &pyobj_attr))
         return(NULL);
-    ctxt = (xsltTransformContextPtr) PytransformCtxt_Get(pyobj_ctxt);
+    style = (xsltStylesheetPtr) Pystylesheet_Get(pyobj_style);
+    attr = (xmlAttrPtr) PyxmlNode_Get(pyobj_attr);
 
-    c_retval = ctxt->style;
-    py_retval = libxslt_xsltStylesheetPtrWrap((xsltStylesheetPtr) c_retval);
-    return(py_retval);
+    xsltCompileAttr(style, attr);
+    Py_INCREF(Py_None);
+    return(Py_None);
 }
 
 PyObject *
@@ -1762,18 +1812,6 @@ libxslt_xsltStylesheetGetDoctypeSystem(PyObject *self ATTRIBUTE_UNUSED, PyObject
     return(py_retval);
 }
 
-PyObject *
-libxslt_xsltCalibrateAdjust(PyObject *self ATTRIBUTE_UNUSED, PyObject *args) {
-    long delta;
-
-    if (!PyArg_ParseTuple(args, (char *)"i:xsltCalibrateAdjust", &delta))
-        return(NULL);
-
-    xsltCalibrateAdjust(delta);
-    Py_INCREF(Py_None);
-    return(Py_None);
-}
-
 PyObject *
 libxslt_xsltDebug(PyObject *self ATTRIBUTE_UNUSED, PyObject *args) {
     xsltTransformContextPtr ctxt;
